/*
 * Copyright 2017 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.diirt.datasource.internal;

import org.apache.commons.lang3.builder.ToStringBuilder;
import org.apache.commons.lang3.builder.ToStringStyle;

import com.aquenos.epics.jackie.common.value.ChannelAccessControlsValue;
import com.aquenos.epics.jackie.common.value.ChannelAccessTimeValue;

/**
 * <p>
 * Message payload for the {@link JackieChannelHandler}. This class is used by
 * the channel handler to store a value that was received from a monitor and
 * pass it on to type adapters.
 * </p>
 * 
 * <p>
 * For most data types (except {@code DBR_STRING}), two monitors are created.
 * One that uses the {@code DBR_CTRL_xxx} type and one that uses the
 * {@code DBR_TIME_xxx} type. The first is used for the meta-data and the second
 * is used for the actual value, time-stamp and alarm status. As a consequence,
 * this payload also stores two values - one from the controls monitor and one
 * from the time monitor.
 * </p>
 * 
 * @author Sebastian Marsching
 */
public final class JackieMessagePayload {

    /**
     * <p>
     * Creates a message payload with an exception that was received from the
     * controls monitor. If the {@code oldMessagePayload} is not
     * <code>null</code>, its data is used as a basis for the newly created
     * payload. In particular, its time value or exception is copied to the new
     * payload.
     * </p>
     * 
     * <p>
     * The repeated exception flag is set if the previous payload is specified
     * and has an exception.
     * </p>
     * 
     * @param e
     *            exception reported by the controls monitor.
     * @param oldMessagePayload
     *            previous message payload or <code>null</code> if no such
     *            payload exists.
     * @return new message payload with the specified exception from the
     *         controls monitor.
     */
    public static JackieMessagePayload withControlsException(Throwable e,
            JackieMessagePayload oldMessagePayload) {
        // This method is only used internally, so we use an assertion instead
        // of a real check.
        assert (e != null);
        if (oldMessagePayload == null) {
            return new JackieMessagePayload(e, null, e, false, null, null);
        } else {
            return new JackieMessagePayload(e, null, e,
                    oldMessagePayload.hasException(),
                    oldMessagePayload.timeException,
                    oldMessagePayload.timeValue);
        }
    }

    /**
     * <p>
     * Creates a message payload with a value that was received from the
     * controls monitor. If the {@code oldMessagePayload} is not
     * <code>null</code>, its data is used as a basis for the newly created
     * payload. In particular, its time value or exception is copied to the new
     * payload.
     * </p>
     * 
     * <p>
     * The repeated exception flag is set if the previous payload is specified
     * and has an exception associated with the time monitor.
     * </p>
     *
     * @param value
     *            value received from the controls monitor.
     * @param oldMessagePayload
     *            previous message payload or <code>null</code> if no such
     *            payload exists.
     * @return new message payload with the specified value from the time
     *         monitor.
     */
    public static JackieMessagePayload withControlsValue(
            ChannelAccessControlsValue<?> value,
            JackieMessagePayload oldMessagePayload) {
        // This method is only used internally, so we use an assertion instead
        // of a real check.
        assert (value != null);
        if (oldMessagePayload == null) {
            return new JackieMessagePayload(null, value, null, false, null,
                    null);
        } else {
            return new JackieMessagePayload(null, value,
                    oldMessagePayload.timeException,
                    oldMessagePayload.hasException(),
                    oldMessagePayload.timeException,
                    oldMessagePayload.timeValue);
        }
    }

    /**
     * <p>
     * Creates a message payload with an exception that was received from the
     * time monitor. If the {@code oldMessagePayload} is not <code>null</code>,
     * its data is used as a basis for the newly created payload. In particular,
     * its controls value or exception is copied to the new payload.
     * </p>
     * 
     * <p>
     * The repeated exception flag is set if the previous payload is specified
     * and has an exception.
     * </p>
     * 
     * @param e
     *            exception reported by the time monitor.
     * @param oldMessagePayload
     *            previous message payload or <code>null</code> if no such
     *            payload exists.
     * @return new message payload with the specified exception from the time
     *         monitor.
     */
    public static JackieMessagePayload withTimeException(Throwable e,
            JackieMessagePayload oldMessagePayload) {
        // This method is only used internally, so we use an assertion instead
        // of a real check.
        assert (e != null);
        if (oldMessagePayload == null) {
            return new JackieMessagePayload(null, null, e, false, e, null);
        } else {
            return new JackieMessagePayload(oldMessagePayload.controlsException,
                    oldMessagePayload.controlsValue, e,
                    oldMessagePayload.hasException(), e, null);
        }
    }

    /**
     * <p>
     * Creates a message payload with a value that was received from the time
     * monitor. If the {@code oldMessagePayload} is not <code>null</code>, its
     * data is used as a basis for the newly created payload. In particular, its
     * controls value or exception is copied to the new payload.
     * </p>
     * 
     * <p>
     * The repeated exception flag is set if the previous payload is specified
     * and has an exception associated with the controls monitor.
     * </p>
     *
     * @param value
     *            value received from the time monitor.
     * @param oldMessagePayload
     *            previous message payload or <code>null</code> if no such
     *            payload exists.
     * @return new message payload with the specified value from the time
     *         monitor.
     */
    public static JackieMessagePayload withTimeValue(
            ChannelAccessTimeValue<?> value,
            JackieMessagePayload oldMessagePayload) {
        // This method is only used internally, so we use an assertion instead
        // of a real check.
        assert (value != null);
        if (oldMessagePayload == null) {
            return new JackieMessagePayload(null, null, null, false, null,
                    value);
        } else {
            return new JackieMessagePayload(oldMessagePayload.controlsException,
                    oldMessagePayload.controlsValue,
                    oldMessagePayload.controlsException,
                    oldMessagePayload.hasException(), null, value);
        }
    }

    private final Throwable controlsException;
    private final ChannelAccessControlsValue<?> controlsValue;
    private final Throwable lastException;
    private final boolean previousHadException;
    private final Throwable timeException;
    private final ChannelAccessTimeValue<?> timeValue;

    private JackieMessagePayload(Throwable controlsException,
            ChannelAccessControlsValue<?> controlsValue,
            Throwable lastException, boolean previousHadException,
            Throwable timeException, ChannelAccessTimeValue<?> timeValue) {
        this.controlsException = controlsException;
        this.controlsValue = controlsValue;
        this.lastException = lastException;
        this.previousHadException = previousHadException;
        this.timeException = timeException;
        this.timeValue = timeValue;
    }

    /**
     * Returns the last value received from the controls monitor or
     * <code>null</code> if no such value exists. The reason why no such value
     * exists may be that no event has been received from the controls monitor
     * yet or that there is an exception associated with the controls monitor.
     * 
     * @return last received controls value or <code>null</code> if there is no
     *         such value.
     */
    public ChannelAccessControlsValue<?> getControlsValue() {
        return controlsValue;
    }

    /**
     * Returns the last exception. If there only is an exception associated with
     * the controls monitor, that exception is returned. If there only is an
     * exception associated with the time monitor, that exception is returned.
     * If there are exceptions associated with both monitors, the exception that
     * has been received later is returned. If there is no exception associated
     * with either of the monitors, <code>null</code> is returned.
     * 
     * @return last exception received by a monitor or <code>null</code> if
     *         there is no exception.
     */
    public Throwable getLastException() {
        return lastException;
    }

    /**
     * Returns the last value received from the time monitor or
     * <code>null</code> if no such value exists. The reason why no such value
     * exists may be that no event has been received from the time monitor yet
     * or that there is an exception associated with the time monitor.
     * 
     * @return last received controls value or <code>null</code> if there is no
     *         such value.
     */
    public ChannelAccessTimeValue<?> getTimeValue() {
        return timeValue;
    }

    /**
     * Tells whether there is an exception associated with either of the
     * monitors. This means this method returns <code>true</code> if
     * {@link #getLastException()} returns an exception and <code>false</code>
     * exactly if {@link #getLastException()} returns <code>null</code>.
     * 
     * @return <code>true</code> if there is an exception associated with either
     *         of the monitors, <code>false</code> otherwise.
     */
    public boolean hasException() {
        return lastException != null;
    }

    /**
     * Tells whether this payload has an exception and the previous payload had
     * an exception as well. This means that this method returns
     * <code>true</code> if this payload's {@link #hasException()} method
     * returns <code>true</code> and the {@link #hasException()} method of the
     * previous payload also returned <code>true</code>.
     * 
     * @return <code>true</code> if both this payload and the previous payload
     *         have an exception, <code>false</code> otherwise.
     */
    public boolean isRepeatedException() {
        return hasException() && previousHadException;
    }

    @Override
    public String toString() {
        return new ToStringBuilder(this, ToStringStyle.SHORT_PREFIX_STYLE)
                .append("controlsException", controlsException)
                .append("controlsValue", controlsValue)
                .append("lastException", lastException)
                .append("previousHadException", previousHadException)
                .append("timeException", timeException)
                .append("timeValue", timeValue).build();
    }

}
