/*
 * Copyright 2017 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.diirt.vtype;

import java.nio.ByteBuffer;
import java.nio.charset.Charset;
import java.time.Instant;

import org.apache.commons.lang3.builder.HashCodeBuilder;
import org.diirt.vtype.VString;
import org.diirt.vtype.VTypeToString;

import com.aquenos.epics.jackie.common.util.NullTerminatedStringUtil;
import com.aquenos.epics.jackie.common.value.ChannelAccessTimeChar;

/**
 * {@link VString} implementation that is backed by a
 * {@link ChannelAccessTimeChar}. This implementation is suitable when an array
 * of type {@code DBR_TIME_CHAR} shall be interpreted as a null-terminated
 * string.
 * 
 * @see JackieVString
 * @author Sebastian Marsching
 */
public final class JackieVStringFromByteArray
        extends JackieAlarmAndTime<ChannelAccessTimeChar> implements VString {

    private final Charset charset;

    /**
     * <p>
     * Creates a value backed by the specified {@code value}. The {@code value}
     * is used for the alarm status, time-stamp, and value. The value is
     * interpreted as null-terminated string that is decoded using the specified
     * {@code charset}.
     * </p>
     * 
     * <p>
     * If {@code disconnectTime} is not <code>null</code>, it is used as the
     * time-stamp and the alarm severity is set to {@code UNDEFINED}.
     * </p>
     * 
     * @param value
     *            value used for alarm status, time-stamp, and value.
     * @param disconnectTime
     *            time at which the channel was disconnected. Usually this
     *            should be <code>null</code>. If it is not <code>null</code>,
     *            the alarm severity is set to {@code UNDEFINED} and this
     *            parameter is used as the time-stamp (instead of using the
     *            time-stamp from the {@code timeValue}.
     * @param charset
     *            character set that is used for decoding the array of bytes
     *            into a string.
     */
    public JackieVStringFromByteArray(ChannelAccessTimeChar value,
            Instant disconnectTime, Charset charset) {
        super(value, disconnectTime, false);
        if (charset == null) {
            throw new NullPointerException();
        }
        this.charset = charset;
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == null || !super.equals(obj)) {
            return false;
        }
        // The equals(...) implementation in the super-class already checks
        // that the types match, so we can cast without further tests.
        JackieVStringFromByteArray other = (JackieVStringFromByteArray) obj;
        return (this.charset == null && other.charset == null)
                || (this.charset != null && this.charset.equals(other.charset));
    }

    @Override
    public String getValue() {
        ByteBuffer buffer = timeValue.getValue();
        byte[] bytes;
        if (buffer.hasArray()) {
            bytes = buffer.array();
        } else {
            bytes = new byte[buffer.remaining()];
            buffer.get(bytes);
        }
        return NullTerminatedStringUtil.nullTerminatedBytesToString(bytes,
                charset);
    }

    @Override
    public int hashCode() {
        return new HashCodeBuilder(97, 31).appendSuper(super.hashCode())
                .append(charset).build();
    }

    @Override
    public String toString() {
        return VTypeToString.toString(this);
    }

}
