/*
 * Copyright 2017 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.diirt.vtype.internal;

import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;
import java.text.NumberFormat;
import java.util.Locale;
import java.util.concurrent.ConcurrentHashMap;

/**
 * Utilities for dealing with {@link NumberFormat}s. The utility methods in this
 * class a designed provide the same results, regardless of the {@link Locale}
 * that is currently set. They always use the period (.) as the decimal
 * separator and never use a grouping separator. The string "Infinity" is used
 * as the symbol for infinity and the string "NaN" is used as the symbol for
 * not-a-number.
 * 
 * @author Sebastian Marsching
 */
public final class NonLocalizedNumberFormats {

    private static final DecimalFormatSymbols DECIMAL_FORMAT_SYMBOLS;
    private static final NumberFormat DEFAULT_NUMBER_FORMAT;
    private static final ConcurrentHashMap<Integer, NumberFormat> NUMBER_FORMATS = new ConcurrentHashMap<Integer, NumberFormat>();

    static {
        DECIMAL_FORMAT_SYMBOLS = DecimalFormatSymbols.getInstance(Locale.US);
        DECIMAL_FORMAT_SYMBOLS
                .setInfinity(Double.toString(Double.POSITIVE_INFINITY));
        DECIMAL_FORMAT_SYMBOLS.setNaN(Double.toString(Double.NaN));
        DEFAULT_NUMBER_FORMAT = new DecimalFormat("0.0",
                DECIMAL_FORMAT_SYMBOLS);
        // We determine the number of fractional digits by looking at the number
        // format of the en_US locale.
        NumberFormat usNumberFormat = NumberFormat.getInstance(Locale.US);
        DEFAULT_NUMBER_FORMAT.setMaximumFractionDigits(
                usNumberFormat.getMaximumFractionDigits());
        DEFAULT_NUMBER_FORMAT.setMinimumFractionDigits(
                usNumberFormat.getMinimumFractionDigits());
    }

    /**
     * <p>
     * Returns a number format that formats a number using a default number of
     * fraction digits. The minimum and maximum number of fraction digits is set
     * according to the defaults for the {@link Locale#US US} locale.
     * </p>
     * 
     * <p>
     * The returned format is cloned before returning it, so the called may
     * modify it.
     * </p>
     * 
     * @return number format using a default number of fraction digits.
     */
    public static NumberFormat getDefaultNumberFormat() {
        // NumberFormats are mutable, so we have to clone the object before
        // returning it.
        return (NumberFormat) DEFAULT_NUMBER_FORMAT.clone();
    }

    /**
     * <p>
     * Returns a number format that formats a number using the specified number
     * of fraction digits. If the specified number is less than or equal to
     * zero, the number is formatted without a fractional part.
     * </p>
     * 
     * <p>
     * The returned format is cloned before returning it, so the called may
     * modify it.
     * </p>
     * 
     * @param precision
     *            number of fractional digits that shall be included in the
     *            format. This number specifies the fixed number of fractional
     *            digits that shall be used (both minimum and maximum).
     * @return number format using the specified number of fractional digits.
     */
    public static NumberFormat getNumberFormatForPrecision(int precision) {
        if (precision < 0) {
            precision = 0;
        }
        NumberFormat numberFormat = NUMBER_FORMATS
                .get(Integer.valueOf(precision));
        if (numberFormat == null) {
            numberFormat = createNumberFormat(precision);
            // As constructing a DecimalFormat from a pattern is relatively
            // expensive, we cache created NumberFormats.
            NumberFormat existingNumberFormat = NUMBER_FORMATS
                    .putIfAbsent(Integer.valueOf(precision), numberFormat);
            // We limit the number of cached formats in order to avoid a DoS
            // attack in which a malicious server sends us many different
            // precisions in order to exhaust our memory. The limit of 1024
            // entries is so big that it not be hit in any regular scenario.
            // When we hit the limit, there are two choices: We could stop
            // adding more entries or we can clear existing entries. We choose
            // the latter option because it seems reasonable that NumberFormats
            // created now might rather be reused than NumberFormats that have
            // been created a very long time ago. In theory, an LRU cache would
            // be best, but this would mean putting more effort into this than
            // really makes sense.
            if (NUMBER_FORMATS.size() > 1024) {
                NUMBER_FORMATS.clear();
            }
            if (existingNumberFormat != null) {
                numberFormat = existingNumberFormat;
            }
        }
        // NumberFormats are mutable, so we have to clone the object before
        // returning it.
        return (NumberFormat) numberFormat.clone();
    }

    private static NumberFormat createNumberFormat(int precision) {
        StringBuilder sb = new StringBuilder("0");
        if (precision > 0) {
            sb.append('.');
            for (int i = 0; i < precision; ++i) {
                sb.append('0');
            }
        }
        return new DecimalFormat(sb.toString(), DECIMAL_FORMAT_SYMBOLS);
    }

    // Private constructor - this class only has static method.
    private NonLocalizedNumberFormats() {
    }

}
